/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Graph.cpp                                                                  *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#include "Graph.h"

extern int log_level;

uint32_t networkNodeNum = 0;
uint32_t linkNum = 0;
uint32_t consumptionNodeNum = 0;
uint32_t hardwareLevelNum = 0;

/**
 * @brief maintaining the map from arc ID to the point to arc, cost only O(1).
 *
 * 1. range of links that between network nodes is [0, 2*linkNum), and all integer indices are filled;
 * 2. range of links that between access network node super sink node is [2*linkNum, 2*linkNum+consumptionNodeNum);
 * 3. range of links that between super source node and server deployment nodes is 
 *    [2*linkNum+consumptionNodeNum, 2*linkNum+consumptionNodeNum+networkNodeNum);
 * 4. range of artificial dummy arcs is [2*linkNum+consumptionNodeNum+networkNodeNum, 2*linkNum+2*consumptionNodeNum+2*networkNodeNum+2).
 */
static struct Arc *arcTable;
/** maintaining the map from server hardware level to its price, cost only O(1). */
std::vector<int> priceTable;
/** maintaining the map from server hardware level to its output capacity, cost only O(1). */
std::vector<int> outcapTable;
/** maintaining the map from network node ID to its deployment fee, cost only O(1). */
std::vector<int> deployTable;
/** maintaining the map from comsumption node ID to its access network node ID, cost only O(1). */
std::vector<int> accessTable;
/** maintaining the map from comsumption node ID to its bandwidth demand, cost only O(1). */
std::vector<int> demandTable;
/** maintaining the map from access node ID to its provide comsumption node ID (reverse table of accessTable), cost only O(1). */
std::unordered_map<int /* access node */, int /* consumption node */> provideTable;

/** automatically increase ID for all arcs. */
static int arcIDIncrement = 0;

/** ownID must less than 2*linkNum. */
struct Arc* oppositeArc(int ownID)
{
	if (ownID % 2 == 0)
		return &arcTable[ownID+1]; // 0 to 1
	else
		return &arcTable[ownID-1]; // 1 to 0
}

Arc::Arc() : arcID(arcIDIncrement), srcID(-1), dstID(-1), bandwidth(0), rental(INF), flow(0), status(ArcStatus::L), nextArc(NULL)
{
	++arcIDIncrement;
}

Node::Node() : potential(0), treeLevel(1), restrictedCap(outcapTable[hardwareLevelNum-1]), prev(NULL), next(NULL), enterArc(NULL), dummyArc(NULL), firstArc(NULL), lastArc(NULL)
{

}

/**
 * constructor of class Graph.
 */
Graph::Graph(uint32_t _node_num) : nodeNum(_node_num), arcNum(0), mcfStatus(MCFStatus::Unsolved),
	superSource(_node_num), superSink(_node_num+1), dummyRoot(_node_num+2), sinkArcs(2*linkNum),
	sourceArcs(sinkArcs+consumptionNodeNum), dummyArcs(sourceArcs+networkNodeNum), endDummyArcs(dummyArcs+_node_num+2),
	exchangeOperation(false)
{
	NodeList = new Node[nodeNum + 3]; // plus three, one for super source, one for super sink and one for dummy root
	dist = new int[nodeNum + 2];
	color = new Colortype[nodeNum + 2];
	pi = new int[nodeNum + 2];
	arcTable = new Arc[endDummyArcs]; // endDummyArcs == 2*linkNum + consumptionNodeNum + networkNodeNum + _node_num + 2
	arcRecord = new Arc*[nodeNum + 2];
	fluxThroughNode = new int[nodeNum];
}

/**
 * destructor of class Graph.
 */
Graph::~Graph()
{
	delete []fluxThroughNode;
	delete []arcRecord;
	delete []arcTable;
	delete []pi;
	delete []color;
	delete []dist;
	delete []NodeList;
}

/**
 * construct the graph in the structure of adjacency list.
 */
void Graph::construct(std::list<struct LinkTuple>& linkTuples)
{
	std::list<struct LinkTuple>::iterator iter1 = linkTuples.begin(), iter2;
	uint32_t arcID = 0;
	uint32_t curNode = 0;
	uint32_t rank = 2;
	struct Arc *firstarc = NULL, *parc = NULL, *qarc = NULL;

	firstarc = &arcTable[arcID];
	firstarc->srcID = iter1->src;
	firstarc->dstID = iter1->dst;
	firstarc->bandwidth = iter1->bandwidth;
	firstarc->rental = iter1->rental;
	firstarc->nextArc = NULL;

	curNode = iter1->src; // node index start with 0
	NodeList[curNode].firstArc = firstarc;
	NodeList[curNode].lastArc = firstarc;
	while (true)
	{
		iter2 = iter1; // iter2 is always a forword element of iter1
		++iter2;
		if (iter2 == linkTuples.end())
			break;
		if (iter2->src == iter1->src)
		{
			if (rank % 2 == 0)
			{
				arcID += 2;
				parc = &arcTable[arcID];
				parc->srcID = iter2->src;
				parc->dstID = iter2->dst;
				parc->bandwidth = iter2->bandwidth;
				parc->rental = iter2->rental;
				parc->nextArc = NULL;

				if (rank == 2)
					firstarc->nextArc = parc;
				else
					qarc->nextArc = parc;
				NodeList[curNode].lastArc = parc;
			}
			else
			{
				arcID += 2;
				qarc = &arcTable[arcID];
				qarc->srcID = iter2->src;
				qarc->dstID = iter2->dst;
				qarc->bandwidth = iter2->bandwidth;
				qarc->rental = iter2->rental;
				qarc->nextArc = NULL;

				parc->nextArc = qarc;
				NodeList[curNode].lastArc = qarc;
			}
			rank++;
		}
		else
		{
			rank = 2;
			curNode = iter2->src;

			arcID += 2;
			firstarc = &arcTable[arcID];
			firstarc->srcID = curNode;
			firstarc->dstID = iter2->dst;
			firstarc->bandwidth = iter2->bandwidth;
			firstarc->rental = iter2->rental;
			firstarc->nextArc = NULL;

			NodeList[curNode].firstArc = firstarc;
			NodeList[curNode].lastArc = firstarc;
		}
		++iter1;
	}
}

/**
 * @brief tranform directed graph to be undirected graph.
 */
void Graph::undirectify(std::list<struct LinkTuple>& linkTuples)
{
	std::list<struct LinkTuple>::iterator iter = linkTuples.begin();
	struct Arc *parc = NULL;
	
	uint32_t arcID = 1; // set arc x and x+1 be the same link with opposite direction, x is an even integer
	
	for (; iter != linkTuples.end(); ++iter, arcID += 2)
	{
		uint32_t curNode = iter->dst; // alias

		parc = &arcTable[arcID];
		parc->srcID = curNode;
		parc->dstID = iter->src;
		parc->bandwidth = iter->bandwidth;
		parc->rental = iter->rental;
		parc->nextArc = NULL;

		if (NodeList[curNode].firstArc == NULL)
			NodeList[curNode].firstArc = parc;
		else
			NodeList[curNode].lastArc->nextArc = parc; // append to the last
		NodeList[curNode].lastArc = parc; // update lastArc
	}
}

/**
 * destruct the graph and free memory.
 */
void Graph::destruct()
{
	for (uint32_t i = 0; i < nodeNum; ++i)
	{
		NodeList[i].firstArc = NULL;
		NodeList[i].lastArc = NULL;
	}
}

/**
 * display the graph whose structure is adjacency list.
 */
void Graph::display()
{
	if (log_level < DEBUG_LEVEL)
		return;

	printf("Graph's adjacency list displays: \n");
	for (uint32_t i = 0; i < nodeNum; i++)
	{
		struct Arc *darc = NULL;
		printf("node: [%u]", i);
		for (darc = NodeList[i].firstArc; darc != NULL; darc = darc->nextArc)
			printf(" -> %d(%d,%d)", darc->dstID, darc->bandwidth, darc->rental);
		printf("\n");
	}
}

/** a collection of things to be done before executing minimum cost flow algorithm, this function called only once. */
void Graph::preMinimumCostFlow()
{
	uint32_t i = 0;

	sumFlux = 0;
	for (i = 0; i < consumptionNodeNum; ++i)
		sumFlux += demandTable[i];

	candidateListNum = nodeNum > 1000 ? 50 : 30;
	hotListNum = nodeNum > 1000 ? 10 : 6;

	candidate = new Candidate[hotListNum + candidateListNum + 1];

	uint32_t arcID = dummyArcs;
	struct Arc *parc = &arcTable[arcID++];
	parc->srcID = 0;
	parc->dstID = dummyRoot;
	parc->bandwidth = INF;
	parc->rental = 10000000;
	parc->status = ArcStatus::T;
	NodeList[0].prev = &NodeList[dummyRoot];
	NodeList[0].next = &NodeList[1];
	NodeList[0].dummyArc = parc;
	NodeList[0].enterArc = parc;
	for (i = 1; i < nodeNum+1; ++i, ++arcID) // source nodes or transit node
	{
		parc = &arcTable[arcID];
		parc->srcID = static_cast<int>(i);
		parc->dstID = dummyRoot;
		parc->bandwidth = INF;
		parc->rental = 10000000;
		parc->status = ArcStatus::T;
		NodeList[i].prev = &NodeList[i-1];
		NodeList[i].next = &NodeList[i+1];
		NodeList[i].dummyArc = parc;
		NodeList[i].enterArc = parc;
	}
	NodeList[superSource].potential = 0;
	NodeList[superSource].dummyArc->flow = sumFlux;
	parc = &arcTable[arcID];
	parc->srcID = dummyRoot;
	parc->dstID = superSink;
	parc->bandwidth = INF;
	parc->rental = 10000000;
	parc->status = ArcStatus::T;
	NodeList[superSink].potential = 20000000;
	NodeList[superSink].prev = &NodeList[superSource];
	NodeList[superSink].next = NULL;
	NodeList[superSink].dummyArc = parc;
	NodeList[superSink].enterArc = parc;
	NodeList[superSink].dummyArc->flow = sumFlux;
	NodeList[dummyRoot].potential = 10000000;
	NodeList[dummyRoot].treeLevel = 0;
	NodeList[dummyRoot].prev = NULL;
	NodeList[dummyRoot].next = &NodeList[0];

	// modify arcs from super source node to server deployment nodes, these status
	// will stay constant in addSuperSource() or delSuperSource(), thus only do once here
	for (i = sourceArcs; i < dummyArcs; ++i)
	{
		arcTable[i].srcID = superSource;
		arcTable[i].bandwidth = outcapTable[hardwareLevelNum-1]; // throughput of network nodes are the highest level CDN server's output capacity
		arcTable[i].rental = 0; // access link need no rental fee
	}

	/*
	visited = new bool[nodeNum+2];
	fluxFeeArray = new FluxFee[networkNodeNum*networkNodeNum];
	feeMatrix = new FluxFee*[networkNodeNum]; // a point array, convenient to use form feeMatrix[i][j]
	for (uint32_t k = 0; k < networkNodeNum; ++k)
		feeMatrix[k] = fluxFeeArray + k*networkNodeNum;
	for (int i = 0; i < static_cast<int>(networkNodeNum); ++i)
	{
		for (int j = 0; j < static_cast<int>(networkNodeNum); ++j)
		{
			feeMatrix[i][j].from = i;
			feeMatrix[i][j].to = j;
		}
	}
	*/

	_addSuperSink();
}

/** API function for Greedy. */
void Graph::setExchangeOperation(bool _exchange_operation)
{
	exchangeOperation = _exchange_operation;
}

/** API function for Greedy. */
void Graph::setOutputFluxUpbound(uint32_t nodeIdx, int upbound)
{
	NodeList[nodeIdx].restrictedCap = upbound;
}

/** Network Primal Simplex Algorithm. */
long Graph::minimumCostFlow(std::vector<int>& serversOutputFlux)
{
	long totalCost = 0;

	mcfStatus = MCFStatus::Unsolved;

	uint32_t i = 0;
	struct Arc *enteringArc = NULL, *leavingArc = NULL, *parc = NULL;

	groupNum = (arcNum-1)/candidateListNum + 1;
	groupPos = 0;
	tmpCandidateListNum = 0;

	while (mcfStatus == MCFStatus::Unsolved)
	{
		enteringArc = _executeCandidateListPivot();

		if (enteringArc != NULL)
		{
			int n1 = enteringArc->dstID, n2 = enteringArc->srcID;
			int t = 0, tau = enteringArc->status != ArcStatus::U ? enteringArc->bandwidth - enteringArc->flow : enteringArc->flow;

			if (enteringArc->status != ArcStatus::U)
				std::swap(n1, n2);

			leavingArc = NULL;
			int storeN1 = n1, storeN2 = n2;
			bool isLeave = false, isLeavingReductFlow = _reductCost(enteringArc) > 0;
			while (n1 != n2)
			{
				int oldN1 = n1, oldN2 = n2;
				if (NodeList[n1].treeLevel > NodeList[n2].treeLevel)
				{
					parc = NodeList[n1].enterArc;
					t = parc->srcID == n1 ? parc->flow : parc->bandwidth - parc->flow;
					isLeave = parc->srcID == n1;

					n1 = _parent(n1, parc);
				}
				else
				{
					parc = NodeList[n2].enterArc;
					t = parc->srcID != n2 ? parc->flow : parc->bandwidth - parc->flow;
					isLeave = parc->srcID != n2;

					n2 = _parent(n2, parc);
				}
				if (t < tau || (t <= tau && NodeList[oldN1].treeLevel <= NodeList[oldN2].treeLevel))
				{
					tau = t;
					leavingArc = parc;
					isLeavingReductFlow = isLeave;
				}
			}

			if (leavingArc == NULL)
				leavingArc = enteringArc;

			if (tau >= INF)
			{
				mcfStatus = MCFStatus::Unbounded;
				break;
			}

			n1 = storeN1;
			n2 = storeN2;
			if (tau != 0)
			{
				enteringArc->flow += (enteringArc->srcID != n1) ? -tau : tau;

				while (n1 != n2)
				{
					if (NodeList[n1].treeLevel > NodeList[n2].treeLevel)
					{
						parc = NodeList[n1].enterArc;
						parc->flow += (parc->srcID == n1) ? -tau : tau;

						n1 = _parent(n1, NodeList[n1].enterArc);
					}
					else
					{
						parc = NodeList[n2].enterArc;
						parc->flow += (parc->srcID != n2) ? -tau : tau;

						n2 = _parent(n2, NodeList[n2].enterArc);
					}
				}
			}

			if (enteringArc != leavingArc)
			{
				bool isLeavingBringFlow = isLeavingReductFlow == (NodeList[leavingArc->srcID].treeLevel > NodeList[leavingArc->dstID].treeLevel);
				bool isEnteringEqualsN1 = enteringArc->srcID == storeN1;

				n1 = enteringArc->dstID;
				n2 = enteringArc->srcID;

				if (isEnteringEqualsN1 != isLeavingBringFlow)
					std::swap(n1, n2);
			}

			leavingArc->status = isLeavingReductFlow ? ArcStatus::L : ArcStatus::U;

			if (leavingArc != enteringArc)
			{
				enteringArc->status = ArcStatus::T;

				if (NodeList[leavingArc->srcID].treeLevel < NodeList[leavingArc->dstID].treeLevel)
					_updateTree(enteringArc, leavingArc->dstID, n1, n2);
				else
					_updateTree(enteringArc, leavingArc->srcID, n1, n2);

				n2 = enteringArc->dstID;
				int delta = _reductCost(enteringArc);
				if (NodeList[enteringArc->srcID].treeLevel > NodeList[enteringArc->dstID].treeLevel)
				{
					delta = -delta;
					n2 = enteringArc->srcID;
				}

				// execute add potential operation
				int level = NodeList[n2].treeLevel;
				struct Node *pnode = &NodeList[n2];

				do {
					pnode->potential += delta;
					pnode = pnode->next;
				}
				while (pnode != NULL && pnode->treeLevel > level);
			}
		}
		else
		{
			mcfStatus = MCFStatus::Solved;
			for (i = dummyArcs; i < endDummyArcs; ++i)
				if (arcTable[i].flow > 0)
					mcfStatus = MCFStatus::Unfeasible;
		}
	}

	if (mcfStatus == MCFStatus::Solved)
	{
		for (i = sourceArcs; i < arcNum; ++i)
			serversOutputFlux[i-sourceArcs] = arcTable[i].flow; // adapt overlaying method other than clear() and push_back() method for efficiency
		
		if (exchangeOperation)
		{
			memset(fluxThroughNode, 0, networkNodeNum*sizeof(int));
			for (i = 0; i < networkNodeNum; ++i)
				for (parc = NodeList[i].firstArc; parc != NULL; parc = parc->nextArc)
					fluxThroughNode[i] += parc->flow;
		}
	}

	if (mcfStatus == MCFStatus::Solved)
	{
		for (i = 0; i < sourceArcs; ++i)
			if (arcTable[i].status == ArcStatus::T || arcTable[i].status == ArcStatus::U)
				totalCost += arcTable[i].flow * arcTable[i].rental;
	}
	else
		totalCost = -1;

	return totalCost;
}

struct Arc* Graph::_executeCandidateListPivot()
{
	uint32_t i = 0, next = 0, minimeValue = (hotListNum < tmpCandidateListNum) ? hotListNum : tmpCandidateListNum;
	int reductC = 0;
	struct Arc *parc = NULL;

	for (i = 2; i <= minimeValue; ++i)
	{
		parc = candidate[i].arc;
		reductC = _reductCost(parc);

		if ((reductC < 0 && parc->status == ArcStatus::L) || (reductC > 0 && parc->status == ArcStatus::U))
		{
			++next;
			candidate[next].arc = parc;
			candidate[next].absReductC = (reductC >= 0) ? reductC : -reductC;
		}
	}

	tmpCandidateListNum = next;
	uint32_t oldGroupPos = groupPos;
	do {
		for (i = groupPos; i < arcNum; i += groupNum)
		{
			parc = &arcTable[i];

			if (parc->status == ArcStatus::L)
			{
				reductC = _reductCost(parc);
				if (reductC < 0)
				{
					++tmpCandidateListNum;
					candidate[tmpCandidateListNum].arc = parc;
					candidate[tmpCandidateListNum].absReductC = (reductC >= 0) ? reductC : -reductC;
				}
			}
			else if (parc->status == ArcStatus::U)
			{
				reductC = _reductCost(parc);
				if (reductC > 0)
				{
					++tmpCandidateListNum;
					candidate[tmpCandidateListNum].arc = parc;
					candidate[tmpCandidateListNum].absReductC = (reductC >= 0) ? reductC : -reductC;
				}
			}
		}

		if (++groupPos >= groupNum)
			groupPos = 0;
	}
	while (tmpCandidateListNum < hotListNum && groupPos != oldGroupPos);

	if (tmpCandidateListNum > 0)
		_sortCandidateList(1, tmpCandidateListNum);
	
	return tmpCandidateListNum > 0 ? candidate[1].arc : NULL;
}

void Graph::_sortCandidateList(const uint32_t min , const uint32_t max)
{
	uint32_t low = min, high = max;

	int cut = candidate[(low + high)/2].absReductC; // alias
	do {
		while (candidate[low].absReductC > cut) 
			++low;
		while (candidate[high].absReductC < cut) 
			--high;

		if (low < high)
			std::swap(candidate[low], candidate[high]);

		if (low <= high)
		{
			++low;
			--high;
		}
	}
	while ( low <= high );

	if (min < high) 
		_sortCandidateList(min, high);
	if (low < max && low <= hotListNum) 
		_sortCandidateList(low, max);
}

int Graph::_reductCost(struct Arc *parc)
{
	return parc->rental + NodeList[parc->srcID].potential - NodeList[parc->dstID].potential;
}

int Graph::_parent(int index, struct Arc *parc)
{
	if (parc == NULL)
	{
		error_log("parc is NULL when _parent() is called.\n");
		return -1;
	}

	return (parc->srcID == index) ? parc->dstID : parc->srcID;
}

void Graph::_updateTree(struct Arc *enteringArc, const int n, const int n1, const int n2)
{
	int delta = NodeList[n1].treeLevel + 1 - NodeList[n2].treeLevel;
	int root = n2, parent = 0;

	struct Node *prevNode = &NodeList[n1], *lastNode = NULL;
	struct Arc *parc = enteringArc, *qarc = NULL;

	bool ok = false;
	while (ok == false)
	{
		ok = root == n;

		parent = _parent(root, NodeList[root].enterArc);
		lastNode = _cutAndUpdateSubtree(&NodeList[root], delta);
		delta += 2;
		_pasteSubtree(&NodeList[root], lastNode, prevNode);
		prevNode = lastNode;

		qarc = NodeList[root].enterArc;
		NodeList[root].enterArc = parc;
		parc = qarc;
		root = parent;
	}
}

struct Node* Graph::_cutAndUpdateSubtree(struct Node *root, const int delta)
{
	int level = root->treeLevel;
	struct Node *pnode = root;
	while (pnode->next && pnode->next->treeLevel > level)
	{
		pnode = pnode->next;
		pnode->treeLevel += delta;
	}

	root->treeLevel += delta;

	if (root->prev != NULL)
		root->prev->next = pnode->next;
	if (pnode->next != NULL)
		pnode->next->prev = root->prev;

	return pnode;
}

void Graph::_pasteSubtree(struct Node *root, struct Node *lastNode, struct Node *prevNode)
{
	struct Node *nextNode = prevNode->next;
	root->prev = prevNode;
	prevNode->next = root;
	lastNode->next = nextNode;
	if (nextNode != NULL)
		nextNode->prev = lastNode;
}

/** a collection of things to be done after executing minimum cost flow algorithm, this function called only once. */
void Graph::postMinimumCostFlow()
{
	_delSuperSink();

	delete []candidate;

	/*
	delete []visited;
	delete []feeMatrix;
	delete []fluxFeeArray;
	*/
}

/** used to record flow paths after minimum cost maximum flow algorithm finished, this function should be called only once. */
void Graph::recordFlowPath()
{
	int source = superSource;
	int sink = superSink;
	std::vector<int> path;
	path.reserve(16);
	while ( _DFS(source, sink, path, INF) )
		;
}

/** called by recordFlowPath(). */
bool Graph::_DFS(int s, int t, std::vector<int>& path, int tmpFlow)
{
	if (s == t)
	{
		path.push_back(tmpFlow);
		fluxPathList.push_back(path);
		path.clear();
		minFlow = tmpFlow;
		return true;
	}
	struct Arc *parc = NodeList[s].firstArc;
	while (parc != NULL)
	{
		if (parc->flow > 0)
		{
			if (parc->dstID != t)
				path.push_back(parc->dstID);
			if ( _DFS(parc->dstID, t, path, std::min(tmpFlow, parc->flow)) )
			{
				parc->flow -= minFlow;
				return true;
			}
		}
		parc = parc->nextArc;
	}
	return false;
}

/** add the super source node, this function should be called every time right before algorithm starts. */
void Graph::addSuperSource(std::list<int>& servers)
{
	struct Arc *parc = NULL;

	sourceNum = static_cast<uint32_t>(servers.size());
	arcNum = sourceArcs + sourceNum;

	uint32_t arcID = sourceArcs;

	for (std::list<int>::iterator iter = servers.begin(); iter != servers.end(); ++iter, ++arcID)
	{
		parc = &arcTable[arcID];
		parc->dstID = *iter; // network node which deploys a CDN server
		parc->bandwidth = NodeList[parc->dstID].restrictedCap;
		parc->nextArc = NULL;

		if (NodeList[superSource].firstArc == NULL)
			NodeList[superSource].firstArc = parc;
		else
			NodeList[superSource].lastArc->nextArc = parc; // append to the last
		NodeList[superSource].lastArc = parc; // update lastArc too
	}
}

/** delete the super source node, this function should be called every time right after minimumCostFlow() is called. */
void Graph::delSuperSource()
{
	_revertStatusToInitialization();

	NodeList[superSource].firstArc = NULL;
	NodeList[superSource].lastArc = NULL;

	for (uint32_t i = sourceArcs; i < sourceArcs+sourceNum; ++i)
		arcTable[i].dstID = -1;
}

/** display arcs from super source to deployment node, only useful when debugging. */
void Graph::displaySuperSource()
{
	if (log_level < DEBUG_LEVEL)
		return;

	printf("Super source node's adjacency list displays: \n");

	struct Arc *darc = NULL;
	printf("node: [%d]", superSource);
	for (darc = NodeList[superSource].firstArc; darc != NULL; darc = darc->nextArc)
		printf(" -> %d(%d,%d)", darc->dstID, darc->bandwidth, darc->rental);
	printf("\n");
}

/** revert the status of all arcs to initialization, this function automatically called by delSuperSource(). */
void Graph::_revertStatusToInitialization()
{
	uint32_t i;
	// revert arcs' status
	for (i = 0; i < arcNum; ++i)
	{
		arcTable[i].flow = 0;
		arcTable[i].status = ArcStatus::L;
	}
	for (i = dummyArcs; i < endDummyArcs; ++i)
	{
		arcTable[i].flow = 0;
		arcTable[i].status = ArcStatus::T;
	}
	arcTable[endDummyArcs-2].flow = sumFlux;
	arcTable[endDummyArcs-1].flow = sumFlux;

	// revert nodes' status
	NodeList[0].potential = 0;
	NodeList[0].treeLevel = 1;
	NodeList[0].prev = &NodeList[dummyRoot];
	NodeList[0].next = &NodeList[1];
	NodeList[0].enterArc = NodeList[0].dummyArc;
	for (i = 1; i < nodeNum+1; ++i) // source nodes or transit node
	{
		NodeList[i].potential = 0;
		NodeList[i].treeLevel = 1;
		NodeList[i].prev = &NodeList[i-1];
		NodeList[i].next = &NodeList[i+1];
		NodeList[i].enterArc = NodeList[i].dummyArc;
	}
	NodeList[superSink].potential = 20000000;
	NodeList[superSink].treeLevel = 1;
	NodeList[superSink].prev = &NodeList[superSource];
	NodeList[superSink].next = NULL;
	NodeList[superSink].enterArc = NodeList[superSink].dummyArc;
	NodeList[dummyRoot].potential = 10000000;
	NodeList[dummyRoot].treeLevel = 0;
	NodeList[dummyRoot].prev = NULL;
	NodeList[dummyRoot].next = &NodeList[0];

	// revert candidates' status
	for (i = 0; i <= hotListNum + candidateListNum; ++i)
	{
		candidate[i].absReductC = 0;
		candidate[i].arc = NULL;
	}
}

/** add the super sink node, this function is called by preMinimumCostFlow(). */
void Graph::_addSuperSink()
{
	struct Arc *parc = NULL;

	uint32_t arcID = sinkArcs;

	for (uint32_t i = 0; i < consumptionNodeNum; ++i, ++arcID)
	{
		int curNode = accessTable[i]; // access node index

		parc = &arcTable[arcID];
		parc->srcID = curNode;
		parc->dstID = superSink;          // super sink node
		parc->bandwidth = demandTable[i]; // demand bandwidth of each consumption node
		parc->rental = 0;                 // access link need no rental fee
		parc->nextArc = NULL;

		if (NodeList[curNode].firstArc == NULL)
			NodeList[curNode].firstArc = parc;
		else
			NodeList[curNode].lastArc->nextArc = parc; // append to the last
		NodeList[curNode].lastArc = parc;  // update lastArc too
	}
}

/** delete the super sink node, this function is called by postMinimumCostFlow(). */
void Graph::_delSuperSink()
{
	uint32_t i = 0;
	for (i = 0; i < consumptionNodeNum; ++i)
	{
		int curNode = accessTable[i];

		struct Arc *lastarc = NodeList[curNode].firstArc, *prevarc = NULL;
		while (lastarc->nextArc != NULL)
		{
			prevarc = lastarc;
			lastarc = lastarc->nextArc;
		}

		if (NodeList[curNode].firstArc->nextArc == NULL)
		{
			NodeList[curNode].firstArc = NULL;
			NodeList[curNode].lastArc = NULL;
		}
		else
		{
			prevarc->nextArc = NULL;
			NodeList[curNode].lastArc = prevarc;
		}
	}

	for (i = sinkArcs; i < sourceArcs; ++i)
		arcTable[i].srcID = -1;
}

/**
 * use BFS algorithm to traverse the graph at the start node src.
 */
void Graph::BFS(int src)
{
	int u, v;
	for (uint32_t i = 0; i < nodeNum; ++i)
	{
		color[i] = Colortype::White;
		dist[i] = INF;
		pi[i] = -1;
	}
	color[src] = Colortype::Gray;
	dist[src] = 0;
	pi[src] = -1;
	arcRecord[src] = NULL;

	std::queue<int> Q;
	Q.push(src);
	while (!Q.empty())
	{
		u = Q.front();
		Q.pop();
		struct Arc *parc = NodeList[u].firstArc;
		while (parc != NULL)
		{
			v = parc->dstID;
			if (color[v] == Colortype::White && parc->rental != INF)
			{
				color[v] = Colortype::Gray;
				dist[v] = dist[u] + 1;
				pi[v] = u;
				arcRecord[v] = parc;
				Q.push(v);
			}
			parc = parc->nextArc;
		}
		color[u] = Colortype::Black;
	}
}

/**
 * use dijkstra algorithm to calculate the minimum distance from node src to other nodes.
 */
void Graph::dijkstra(int src)
{
	int u, v;
	for (uint32_t i = 0; i < nodeNum; ++i)
	{
		dist[i] = INF;
		pi[i] = 0;
	}
	dist[src] = 0;
	pi[src] = -1;
	arcRecord[src] = NULL;

	std::priority_queue<std::pair<int, int>, std::vector<std::pair<int, int> >, std::greater<std::pair<int, int> > > priorityQ;
	priorityQ.push(std::pair<int, int>(0, src));
	while (!priorityQ.empty())
	{
		u = priorityQ.top().second;
		priorityQ.pop();
		struct Arc *parc = NodeList[u].firstArc;
		while (parc != NULL)
		{
			v = parc->dstID;
			if (dist[v] > dist[u] + parc->rental)
			{
				dist[v] = dist[u] + parc->rental;
				pi[v] = u;
				arcRecord[v] = parc;
				priorityQ.push(std::pair<int, int>(dist[v], v));
			}
			parc = parc->nextArc;
		}
	}
}

/**
 * obtain the node path from src to dst after call BFS() or dijkstra().
 */
void Graph::recordNodePath(std::vector<int>& path, int src, int dst)
{
	std::stack<int> S;
	while (dst != src)
	{
		S.push(dst);
		dst = pi[dst];
	}
	S.push(dst);
	while (!S.empty())
	{
		path.push_back(S.top());
		S.pop();
	}
}

/**
 * obtain the arc path from src to dst after call BFS() or dijkstra().
 */
long Graph::recordArcPath(std::vector<struct Arc*>& path, int src, int dst)
{
	std::stack<int> S;
	long costSum = 0;
	while (dst != src)
	{
		int tmp = dst;
		dst = arcRecord[dst]->arcID;
		S.push(dst);
		dst = pi[tmp];
	}
	while (!S.empty())
	{
		path.push_back(&arcTable[S.top()]);
		costSum += arcTable[S.top()].rental;
		S.pop();
	}
	return costSum;
}

/** verification of the final CDN server deployment scheme. */
bool verifyScheme(Graph& graph, long& totalRentalFee, long& totalPurchaseFee)
{
	if (graph.fluxPathList.size() > 300000)
	{
		error_log("The number of flux paths has overstep the limits!\n");
		return false;
	}

	uint32_t i = 0;
	totalRentalFee = 0;
	totalPurchaseFee = 0;
	std::vector<int> arcLeftBandwidth(graph.arcNum, 0);
	std::vector<int> demandSatified(consumptionNodeNum, 0);
	std::unordered_set<int> serverNodeSet;
	for (i = 0; i < graph.arcNum; ++i)
		arcLeftBandwidth[i] = arcTable[i].bandwidth;

	int fluxPathIdx = 1;
	for (std::list<std::vector<int> >::iterator iter = graph.fluxPathList.begin(); iter != graph.fluxPathList.end(); ++iter, ++fluxPathIdx)
	{
		debug_log("verify flux path %d...\n", fluxPathIdx);
		
		std::vector<int>& fluxPath = *iter; // alias
		if (fluxPath.size() < 4)
		{
			error_log("The number of nodes in a flux path less than 3, it must be an invalid flux path!\n");
			return false;
		}
		if (fluxPath.size() > 10000)
		{
			error_log("The number of nodes in a flux path has overstep the limits!\n");
			return false;
		}

		int serverNodeID = fluxPath[0];
		uint32_t accessNodeIdx = static_cast<uint32_t>(fluxPath.size()) - 4;
		int consumptionNodeID = fluxPath[accessNodeIdx+1];
		int consumptionBandwidth = fluxPath[accessNodeIdx+2];
		int serverHardwareLevel = fluxPath.back();
		if (fluxPath[accessNodeIdx] != accessTable[consumptionNodeID])
		{
			error_log("The access network node[%d] in the flux path is not the same with that required[%d]!\n", fluxPath[accessNodeIdx], accessTable[consumptionNodeID]);
			return false;
		}
		demandSatified[consumptionNodeID] += consumptionBandwidth;

		if (serverNodeSet.find(serverNodeID) == serverNodeSet.end())
		{
			serverNodeSet.insert(serverNodeID);
			totalPurchaseFee += priceTable[serverHardwareLevel];
		}

		struct Arc *parc = graph.NodeList[serverNodeID].firstArc;
		for (i = 1; i <= accessNodeIdx; ++i)
		{
			while (parc != NULL)
			{
				if (parc->dstID == fluxPath[i])
				{
					totalRentalFee += parc->rental*consumptionBandwidth;
					arcLeftBandwidth[parc->arcID] -= consumptionBandwidth;
					if (arcLeftBandwidth[parc->arcID] < 0)
					{
						error_log("The left bandwidth of arc[%d] which is %d -> %d belows zero!\n", parc->arcID, parc->srcID, parc->dstID);
						return false;
					}
					parc = graph.NodeList[parc->dstID].firstArc;
					break;
				}
				parc = parc->nextArc;
			}
			if (parc == NULL)
			{
				error_log("The flux path has an arc that is not in the graph!\n");
				return false;
			}
		}
	}
	
	for (uint32_t k = 0; k < consumptionNodeNum; ++k)
	{
		if (demandSatified[k] < demandTable[k])
		{
			error_log("The flux scheme cannot satify the bandwidth requirement of consumption node[%u]!\n", k);
			return false;
		}
	}
	
	return true;
}
