/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Graph.h                                                                    *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#ifndef __GRAPH_H__
#define __GRAPH_H__

#include "Utils.h"

#define INF    0x7fffffff

enum class ArcStatus {
	T, ///< in T, 0 < arc.flow < bandwidth
	L, ///< in L, arc.flow == 0
	U  ///< in U, arc.flow == bandwidth
};

enum class MCFStatus {
	Unsolved,   ///< optimal solution has not be found now.
	Solved,     ///< has found the optimal solution of problem.
	Unfeasible, ///< problem is unfeasible.
	Unbounded   ///< problem is unbounded.
};

/** arc structure of graph. */
struct Arc
{
	Arc();

	int arcID;           ///< ID of this arc.
	int srcID;           ///< tail of this arc.
	int dstID;           ///< head of this arc.
	int bandwidth;       ///< max capacity of this arc.
	int rental;          ///< cost weight of this arc.
	int flow;            ///< initial flow of this arc.
	ArcStatus status;    ///< tells if arc is deleted, closed, in T, L, or U.

	struct Arc *nextArc; ///< next out degree arc of the node this arc derived.
};

/** node structure of graph. */
struct Node
{
	Node();

	int potential;     ///< the node potential corresponding with the flow conservation constrait of this node.
	int treeLevel;     ///< the depth of the node in T as to T root.
	int restrictedCap; ///< restricted output capacity due to imposed server's hardware level limitation.

	struct Node *prev; ///< previous node in the order of the Post-Visit on T.
	struct Node *next; ///< next node in the order of the Post-Visit on T.

	struct Arc *enterArc; ///< entering tree arc of this node.
	struct Arc *dummyArc; ///< artificial arcs are used to connect nodes with the dummy root node.

	struct Arc *firstArc; ///< first out degree arc of this node.
	struct Arc *lastArc;  ///< last out degree arc of this node.
};

/** candidate structure used in minimum cost flow algorithm. */
struct Candidate
{
	Candidate() : absReductC(0), arc(NULL) {}

	int absReductC;  ///< absolute value of the arc's reduced cost.
	struct Arc *arc; ///< pointer to the violating primal bound arc.
};

/** adjacency list structure representing the graph. */
class Graph
{
public:
	explicit Graph(uint32_t _node_num);
	~Graph();

	/** @name regular methods of Graph. */
	///@{
	void construct(std::list<struct LinkTuple>& linkTuples);
	void undirectify(std::list<struct LinkTuple>& linkTuples);
	void destruct();
	void display();
	///@}

	/** @name API function members used by minimum cost flow algorithm. */
	///@{
	void addSuperSource(std::list<int>& servers);
	void delSuperSource();
	void displaySuperSource();
	void preMinimumCostFlow();
	long minimumCostFlow(std::vector<int>& serversOutputFlux);
	void postMinimumCostFlow();
	void recordFlowPath();
	///@}

	/** @name API member functions used by TopologyScanner. */
	///@{
	void BFS(int src);
	void dijkstra(int src);
	void recordNodePath(std::vector<int>& path, int src, int dst);
	long recordArcPath(std::vector<struct Arc*>& path, int src, int dst);
	///@}

public:
	/** struct used by exchange operation in Greedy algorithm. */
	class FluxFee
	{
	public:
		int from;
		int to;
		int fee;

		bool operator<(FluxFee rhs) const { return fee < rhs.fee; }
	};

	/** @name API member functions used by Greedy. */
	///@{
	void setExchangeOperation(bool _exchange_operation);
	void setOutputFluxUpbound(uint32_t nodeIdx, int upbound);
	///@}

private:
	enum Colortype
	{
		White = 0, ///< denote this node has not been visited.
		Gray,      ///< denote this node has been visited but its all out degrees hasn't been visited.
		Black      ///< denote this node's all out degrees has been visited.
	};

	Graph(const Graph&);
	Graph& operator=(const Graph&);

	/** @name internal member functions used by minimum cost flow algorithm. */
	///@{
	struct Arc* _executeCandidateListPivot();
	void _sortCandidateList(const uint32_t min, const uint32_t max);
	int _reductCost(struct Arc *parc);
	int _parent(int index, struct Arc *parc);
	void _updateTree(struct Arc *k, const int h, const int k1, const int k2);
	struct Node* _cutAndUpdateSubtree(struct Node *root, const int delta);
	void _pasteSubtree(struct Node *root, struct Node *lastNode, struct Node *prevNode);

	void _addSuperSink();
	void _delSuperSink();
	void _revertStatusToInitialization();
	bool _DFS(int s, int t, std::vector<int>& path, int tmpFlow);
	///@}

public:
	uint32_t nodeNum;       ///< maintaining the node number of the graph.
	uint32_t arcNum;        ///< maintaining the arc number of the graph.
	struct Node *NodeList;  ///< storing all the nodes in the graph.
	int *dist;              ///< recording the time when a node is discovered.
	int *fluxThroughNode;   ///< used by Greedy in exchange operation.
	std::priority_queue<FluxFee> fluxFeeQueue; ///< used by Greedy.
	std::list<std::vector<int> > fluxPathList; ///< stores all flux paths from super source to super sink, which is certainly the result of this program.

private:
	enum Colortype *color;  ///< recording a node's visited status when BFS or DFS traverse is called.
	int *pi;                ///< maintaining relationship of node's predecessor when BFS, DFS or dijkstra traverse is called.
	struct Arc **arcRecord; ///< maintaining relationship of arc's predecessor when to call recordArcPath method after BFS or dijkstra traverse is called.

	/** @name data members used by minimum cost maximum flow algorithm. */
	///@{
	MCFStatus mcfStatus;          ///< maintaining MCF status

	const uint32_t superSource;   ///< the index of super source node, which is nodeNum.
	const uint32_t superSink;     ///< the index of super sink node, which is nodeNum + 1.
	const uint32_t dummyRoot;     ///< the index of dummy root node, which is nodeNum + 2.
	const uint32_t sinkArcs;      ///< start index of arcs pointing to super sink node.
	const uint32_t sourceArcs;    ///< start index of arcs pointing from super source node.
	const uint32_t dummyArcs;     ///< start index of artificial dummy arcs.
	const uint32_t endDummyArcs;  ///< beyond last index of artificial dummy arcs.

	struct Candidate *candidate;  ///< every element points to an element of the arcs vector which contains an arc violating dual bound
	uint32_t groupNum;            ///< number of the candidate lists
	uint32_t groupPos;            ///< contains the actual candidate list
	uint32_t tmpCandidateListNum; ///< hot list dimension (it is variable)
	uint32_t candidateListNum;    ///< number of candidate lists
	uint32_t hotListNum;          ///< number of candidate lists and hot list dimension

	bool exchangeOperation; ///< indicate whether the exchange operation in greedy algorithm has begun.
	FluxFee *fluxFeeArray;  ///< used by Greedy.
	FluxFee **feeMatrix;    ///< used by Greedy.
	int sumFlux;            ///< sum flux in graph, which is equals to sum up demand flux of all consumption node.
	int minFlow;            ///< used by _DFS function as a global variable.
	uint32_t sourceNum;     ///< stores the number of chosen nodes deployed with servers to maintain arcNum of graph.
	///@}
};

bool verifyScheme(Graph& graph, long& totalRentalFee, long& totalPurchaseFee);

#endif /* __GRAPH_H__ */
