/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Greedy.h                                                                   *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#ifndef __GREEDY_H__
#define __GREEDY_H__

#include "TopologyScanner.h"

#define uuid(n1, n2)    (n1 << 16) + n2

/** Greedy algorithm implemented by a finite state machine(FSM). */
class Greedy
{
public:
	explicit Greedy(TopologyScanner& _scanner);
	~Greedy();

	Greedy(const Greedy& rhs) = delete;
	Greedy& operator=(const Greedy& rhs) = delete;

	void initialize(std::list<int>& servers);
	bool optimize(std::list<int>& servers, std::vector<int>& serversOutputFlux, long justResultFee, long purchaseFee, long deploymentFee);

private:
	enum Progress {
		ZOM,     ///< Zero Order Minus.
		ZOP,     ///< Zero Order Plus.
		DL,      ///< Decrease Level Operation.
		DLSTEP,  ///< Single Step Of Decrease Level Operation.
		IL,      ///< Increase Level Operation.
		ILSTEP,  ///< Single Step Of Increase Level Operation.
		EX,      ///< Exchange Operation.(neighbor)
		EXSTEP,  ///< Single Step Of Exchange Operation.(neighbor)
		EX2,     ///< Exchange Operation.(good)
		EX2STEP, ///< Single Step Of Exchange Operation.(good)
		RM       ///< Remove Operation.
	};

	struct OverFluxTuple
	{
		OverFluxTuple(double _unitFee, int _nodeID, int _preLevel) : unitFee(_unitFee), nodeID(_nodeID), preLevel(_preLevel) {}

		double unitFee; ///< unit over flux fee.
		int nodeID;     ///< node ID.
		int preLevel;   ///< hardware level before decrease operation.
	};

private:
	TopologyScanner &scanner; ///< reference to topologyScanner

	enum Progress progress;   ///< indicates current optimize progress.

	uint32_t minusPlusTimes;  ///< times of executed zero order minus and zero order plus progress.
	const uint32_t MAX_MINUS_PLUS_TIMES; ///< max allowed times of executing zero order minus and zero order plus progress.
	const int MAX_TRY_TIMES_IN_EX;  ///< max allowed trying times of exchanging a certain pair of nodes.

	int curIdx;          ///< current possible consumption node index.
	uint32_t curNBIdx;   ///< current selected exchangeable neighbor node index.
	uint32_t cspCN;      ///< only count those possible consumption node.
	uint32_t neighborCN; ///< only count those exchangeable neighbor node.

	bool *deployed;      ///< if node x has deployed a server, then deployed[i] is true, otherwise false.
	int *cspRank;        ///< consumption rank used in zero order plus and zero order minus progress.

	bool hasExecutedZOP;       ///< whether ZOP progress has been executed.
	bool exchangeSwitch;       ///< true indicates adapt flow through rule in EX2, false indicates adapt unit flux fee rule in EX2.
	bool inDecreaseProgress;   ///< control which part of code to be executed when state just switched to decrease progress.
	bool inIncreaseProgress;   ///< control which part of code to be executed when state just switched to increase progress.
	bool inExchangeProgress;   ///< control which part of code to be executed when state just switched to exchange progress.
	bool inRemovingProgress;   ///< control which part of code to be executed when state just switched to removing progress.
	int tryDecreaseTimes;      ///< store the number we have tried to decrease server level.
	int tryIncreaseTimes;      ///< store the number we have tried to increase server level.
	int tryExchangeTimes;      ///< store the number we have tried to exchange, it is used to resume curIdx when return to exchange operation.
	uint32_t decreasableIdx;   ///< current index of decreasable node.
	uint32_t decreasableNum;   ///< number of decreasable node.
	uint32_t increasableIdx;   ///< current index of increasable node.
	uint32_t increasableNum;   ///< number of increasable node.
	uint32_t exchangeableIdx;  ///< current index of exchangeable node.
	uint32_t exchangeableCIdx; ///< current index of exchangeable node.
	uint32_t exchangeableNum;  ///< number of exchangeable node.
	uint32_t exchangeableCN;   ///< number of exchangeable node.
	uint32_t removableIdx;     ///< current index of removable node.
	uint32_t removableNum;     ///< number of removable node.
	const uint32_t kDecreasableNum;  ///< const number of decreasable node.
	const uint32_t kIncreasableNum;  ///< const number of increasable node.
	const uint32_t kExchangeableNum; ///< const number of exchangeable node.
	const uint32_t kExchangeableCN;  ///< const number of exchangeable node.
	const uint32_t kRemovableNum;    ///< const number of removable node.

	int lastChangedNode;      ///< store the node in previous minus operation for undo operation.
	long lastResultFee;       ///< store the total fee result by previous operation for comparsion purpose.

	std::vector<double> deltaPrice; ///< delta price of two neighbor level, deltaPrice[i] == priceTable[i+1] - priceTable[i], where i >= 0
	std::vector<std::pair<int, int> > decreasableArray; ///< decreasable node array.
	std::vector<std::pair<int, int> > increasableArray; ///< increasable node array.
	std::vector<int> exchangeableArray;    ///< exchangeable node array.
	std::vector<int> neighborArray;        ///< store exchangeable neighbors.
	std::vector<int> exchangeableCArray;   ///< exchangeable node array.
	std::vector<int> removableArray;       ///< removable node array.
	std::unordered_set<int> serverHashset; ///< store current server deployment nodes.
	std::unordered_map<int /* uuid */, int /* times */> exchangeTimes; ///< store how many times a pair of node was exchanged.
	std::list<int>::iterator itS;          ///< iterate servers list for remove a certain server.
};

#endif /* __GREEDY_H__ */
