/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: TopologyScanner.h                                                          *
 * @author: Xu Le                                                                    *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#ifndef __TOPOLOGYSCANNER_H__
#define __TOPOLOGYSCANNER_H__

#include "Graph.h"

class Greedy; // forward declaration

/** A class undertaking to scan graph topology to find some useful results. */
class TopologyScanner
{
public:
	explicit TopologyScanner(Graph& _graph);
	~TopologyScanner();

	TopologyScanner(const TopologyScanner& rhs) = delete;
	TopologyScanner& operator=(const TopologyScanner& rhs) = delete;

	void scan();
	void recommend();

	void resetRestrictedCap();

private:
	/** description of the density of arcs in graph. */
	enum class ArcDensity {
		Unknown = 0,
		UltraSparse,
		VerySparse,
		Sparse,
		General,
		Dense,
		VeryDense,
		UltraDense
	};
	/** description of the density of consumption nodes in graph. */
	enum class CspDensity {
		Unknown = 8,
		UltraSparse,
		VerySparse,
		Sparse,
		General,
		Dense,
		VeryDense,
		UltraDense
	};
	const char* _arcDensityString();
	const char* _cspDensityString();
	void _defineDensity();
	void _mustDeployOnAccess();
	void _calcCheapestAmongAccess();
	void _scanNeighborOfAccess();
	void _decideMultiplicativeFactor();
	void _selectByScore();

private:
	friend class Greedy;

	Graph &graph; ///< reference to graph.
	ArcDensity arcDensity; ///< stores the density of arcs in graph.
	CspDensity cspDensity; ///< stores the density of consumption nodes in graph.
	int mNetworkNodeNum;   ///< equals to networkNodeNum, just get rid of annoying type incompatible warning.

	/** @name status of all network nodes. */
	///@{
	int *outLinkCap;       ///< the sum up capacity of all links of a network node.
	int *nodeLevel;        ///< the out link capacity level of a network node.
	double *outLinkFee;    ///< the weighted average rental fee of all links of a network node.
	double *capRank;       ///< the sum up capacity rank of a network node.
	double *rentalFeeRank; ///< the weighted average rental fee rank of a network node.
	double *deployFeeRank; ///< the deployment fee rank of a network node.
	int *scoreTable;       ///< stores score of all network nodes.
	double capVariance;    ///< variance of node output capacity
	///@}

	double multiplicativeFactor;  ///< the result multiplied by consumption node num is the number of network nodes to be selected as possible servers' deployment location. (decided by cspDensity)
	std::vector<int> possibleServers; ///< all network nodes with high score that are possibly chosen as where to deploy servers.
	std::set<int> definiteServers;    ///< set of access nodes that have to deploy a cdn server.
};

#endif /* __TOPOLOGYSCANNER_H__ */
