/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Utils.cpp                                                                  *
 * @author: Xu Le                                                                    *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#include "Utils.h"

extern int log_level;

extern int maxUnitRentalFee;

extern uint32_t networkNodeNum;
extern uint32_t linkNum;
extern uint32_t consumptionNodeNum;
extern uint32_t hardwareLevelNum;

extern std::vector<int> priceTable;
extern std::vector<int> outcapTable;
extern std::vector<int> deployTable;
extern std::vector<int> accessTable;
extern std::vector<int> demandTable;
extern std::unordered_map<int, int> provideTable;

void parseInput(char *filename, std::list<struct LinkTuple>& linkTuples)
{
	FILE *fd = fopen(filename, "r");
	if (fd == NULL)
	{
		error_log("Fail to open file %s.\n", filename);
		exit(EXIT_FAILURE);
	}

	uint32_t i = 0;
	int srcNode = 0, dstNode = 0, bandwidth = 0, rental = 0;
	int deployFee = 0;
	int level = 0, outcap = 0, price = 0;
	int cspID = 0, netID = 0, demand = 0;

	fscanf(fd, "%u %u %u\n", &networkNodeNum, &linkNum, &consumptionNodeNum);

	char buf[100] = { '\0' };
	while (true)
	{
		fgets(buf, 100, fd);
		if (buf[0] == '\r' || buf[0] == '\n')
			break;
		sscanf(buf, "%d %d %d", &level, &outcap, &price);
		outcapTable.push_back(outcap);
		priceTable.push_back(price);
		++hardwareLevelNum;
	}

	deployTable.resize(networkNodeNum);
	accessTable.resize(consumptionNodeNum);
	demandTable.resize(consumptionNodeNum);

	for (i = 0; i < networkNodeNum; ++i)
	{
		fscanf(fd, "%d %d", &netID, &deployFee);
		deployTable[netID] = deployFee;
	}

	fgetc(fd);

	for (i = 0; i < linkNum; ++i)
	{
		fscanf(fd, "%d %d %d %d", &srcNode, &dstNode, &bandwidth, &rental);
		linkTuples.emplace_back(srcNode, dstNode, bandwidth, rental);
		if (maxUnitRentalFee < rental)
			maxUnitRentalFee = rental;
	}

	fgetc(fd);

	for (i = 0; i < consumptionNodeNum; ++i)
	{
		fscanf(fd, "%d %d %d", &cspID, &netID, &demand);
		accessTable[cspID] = netID;
		demandTable[cspID] = demand;
	}

	for (i = 0; i < consumptionNodeNum; ++i)
		provideTable.insert(std::pair<int, int>(accessTable[i], i));

	fclose(fd);
}

void outputToFile(char *filename, std::list<std::vector<int> >& fluxPathList)
{
	FILE *fd = fopen(filename, "w");
	if (fd == NULL)
	{
		error_log("Fail to open file %s.\n", filename);
		exit(EXIT_FAILURE);
	}

	fprintf(fd, "%u\n\n", static_cast<uint32_t>(fluxPathList.size()));

	for (std::list<std::vector<int> >::iterator iter = fluxPathList.begin(); iter != fluxPathList.end(); ++iter)
	{
		std::vector<int> &fluxPath = *iter; // alias
		for (size_t i = 0; i < fluxPath.size()-1; ++i)
			fprintf(fd, "%d ", fluxPath[i]);
		fprintf(fd, "%d\n", fluxPath.back());
	}

	fclose(fd);
}

void outputToFile(char *filename)
{
	FILE *fd = fopen(filename, "w");
	if (fd == NULL)
	{
		error_log("Fail to open file %s.\n", filename);
		exit(EXIT_FAILURE);
	}

	fprintf(fd, "NA\n");

	fclose(fd);
}

bool binarySearch(const std::vector<int>& vec, const int key)
{
    int low = 0, high = static_cast<int>(vec.size())-1, mid = 0;
    while (low <= high)
    {
        mid = low + ((high - low) >> 1);
        if (key == vec[mid])
            return true;
        else if (key < vec[mid])
            high = mid - 1;
        else
            low = mid + 1;
    }
    return false;
}

void defineAvailableLevel()
{
	uint32_t i = 0;
	double sixthPosDeployFee = 0.0;
	std::vector<double> unitOutputCapFee(hardwareLevelNum, 0.0);
	std::vector<double> sortDeployFee;
	sortDeployFee.reserve(networkNodeNum);
	for (i = 0; i < networkNodeNum; ++i)
		sortDeployFee.push_back(deployTable[i]);

	std::sort(sortDeployFee.begin(), sortDeployFee.end());

	sixthPosDeployFee = sortDeployFee[networkNodeNum/6];
	info_log("sixth position deploy fee: %f\n", sixthPosDeployFee);

	for (i = 0; i < hardwareLevelNum; ++i)
	{
		unitOutputCapFee[i] = (priceTable[i] + sixthPosDeployFee) / outcapTable[i];
		info_log("level [%u]'s unit output capacity fee is %f\n", i, unitOutputCapFee[i]);
	}

	const uint32_t oldHardwareLevelNum = hardwareLevelNum;
	const uint32_t L = 4;
	if (hardwareLevelNum > L) // 5, 6, 7, 8, 9, 10
	{
		double *deltaUnitCapFee = new double[hardwareLevelNum-L];
		double *deltaFeeRatio = NULL;
		if (hardwareLevelNum > L+1)
			deltaFeeRatio = new double[hardwareLevelNum-L-1];
		for (i = L; i < hardwareLevelNum; ++i)
			deltaUnitCapFee[i-L] = unitOutputCapFee[i] - unitOutputCapFee[i-1];
		for (i = 0; i < hardwareLevelNum-L; ++i)
			info_log("deltaUnitCapFee[%u]: %f\n", i+L-1, deltaUnitCapFee[i]);
		if (hardwareLevelNum > L+1)
		{
			uint32_t maxDeltaIdx = 0;
			double maxDeltaFeeRatio = -100.0;
			for (i = 0; i < hardwareLevelNum-L-1; ++i)
			{
				deltaFeeRatio[i] = deltaUnitCapFee[i+1] - deltaUnitCapFee[i];
				if (deltaFeeRatio[i] < 0)
					deltaFeeRatio[i] = -deltaFeeRatio[i];
				if (maxDeltaFeeRatio < deltaFeeRatio[i])
				{
					maxDeltaFeeRatio = deltaFeeRatio[i];
					maxDeltaIdx = i+L-1;
				}
				info_log("deltaFeeRatio[%u]: %f\n", i+L-1, deltaFeeRatio[i]);
			}
			hardwareLevelNum = maxDeltaIdx + 2;
			if (hardwareLevelNum > oldHardwareLevelNum)
				hardwareLevelNum = oldHardwareLevelNum;
			uncond_log("highest available server level limited to %u\n", hardwareLevelNum-1);
			delete []deltaFeeRatio;
		}
		delete []deltaUnitCapFee;
	}
}

int locateLevel(const int nodeCap)
{
	int low = 0, high = static_cast<int>(hardwareLevelNum), mid = 0;
	while (high - low > 1)
	{
		mid = low + ((high - low) >> 1);
		if (nodeCap < outcapTable[mid])
			high = mid;
		else
			low = mid;
	}
	return low;
}

int fitLevel(const int outputFlux)
{
	if (outputFlux > outcapTable[hardwareLevelNum-1])
		return hardwareLevelNum;

	int low = -1, high = static_cast<int>(hardwareLevelNum)-1, mid = 0;
	while (high - low > 1)
	{
		mid = low + ((high - low) >> 1);
		if (outputFlux <= outcapTable[mid])
			high = mid;
		else
			low = mid;
	}
	return high;
}

long obtainTotalPurchaseFee(std::vector<int>& serversOutputFlux, size_t serverNum)
{
	long totalPurchaseFee = 0;

	for (size_t i = 0; i < serverNum; ++i)
	{
		if (serversOutputFlux[i] > 0)
		{
			int level = fitLevel(serversOutputFlux[i]);
			totalPurchaseFee += priceTable[level];
		}
	}

	return totalPurchaseFee;
}

Complexity defineDefaultComplexity()
{
	if (consumptionNodeNum > 300)
		return Complexity::Hell;
	if (consumptionNodeNum > 200 || networkNodeNum > 800)
		return Complexity::UltraHard;
	if (consumptionNodeNum > 100 || networkNodeNum > 500)
		return Complexity::VeryHard;
	if (consumptionNodeNum > 50 || networkNodeNum > 300)
		return Complexity::Hard;
	if (consumptionNodeNum > 20 || networkNodeNum > 100)
		return Complexity::General;
	if (consumptionNodeNum > 10 || networkNodeNum > 50)
		return Complexity::Simple;
	return Complexity::Oversimplified;
}

const char* complexityString(Complexity complexity)
{
	switch (complexity)
	{
	case Complexity::Oversimplified:
		return "oversimplified";
	case Complexity::Simple:
		return "simple";
	case Complexity::General:
		return "general";
	case Complexity::Hard:
		return "hard";
	case Complexity::VeryHard:
		return "very-hard";
	case Complexity::UltraHard:
		return "ultra-hard";
	default:
		return "hell";
	}
}

////////////////    convenient functions of std::vector and std::list    ////////////////
void printVector(std::vector<int>& _vector, const char *_note)
{
	if (log_level >= INFO_LEVEL)
	{
		printf("%s", _note);
		for (std::vector<int>::iterator iter = _vector.begin(); iter != _vector.end(); ++iter)
			printf("%d ", *iter);
		printf("\n");
	}
}

void printVector(std::vector<double>& _vector, const char *_note)
{
	if (log_level >= INFO_LEVEL)
	{
		printf("%s", _note);
		for (std::vector<double>::iterator iter = _vector.begin(); iter != _vector.end(); ++iter)
			printf("%f ", *iter);
		printf("\n");
	}
}

void printList(std::list<int>& _list, const char *_note)
{
	if (log_level >= INFO_LEVEL)
	{
		printf("%s", _note);
		for (std::list<int>::iterator iter = _list.begin(); iter != _list.end(); ++iter)
			printf("%d ", *iter);
		printf("\n");
	}
}

void printList(std::list<double>& _list, const char *_note)
{
	if (log_level >= INFO_LEVEL)
	{
		printf("%s", _note);
		for (std::list<double>::iterator iter = _list.begin(); iter != _list.end(); ++iter)
			printf("%f ", *iter);
		printf("\n");
	}
}
