/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: main.cpp                                                                   *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#include "Greedy.h"

extern int log_level;

extern uint32_t networkNodeNum;
extern uint32_t linkNum;
extern uint32_t consumptionNodeNum;
extern uint32_t hardwareLevelNum;

extern std::vector<int> priceTable;
extern std::vector<int> outcapTable;
extern std::vector<int> deployTable;
extern std::vector<int> accessTable;
extern std::vector<int> demandTable;
extern std::unordered_map<int, int> provideTable;

std::list<int> curBestServers;

int main(int argc, char *argv[])
{
	Timer procedureTimer("Elapsed Time: ");
	/*******************************************************************
	 ********************     Time elapsing ...     ********************
	 *******************************************************************/
	// printf("current log level: %s\n\n", getLogLevel());

	std::list<struct LinkTuple> linkTuples;

	// 1. ===============    parse input char* array    ===============
	parseInput(argv[1], linkTuples);

	if (networkNodeNum > 1000)
		defineAvailableLevel();
	uint32_t bestRestrictedLevelNum = hardwareLevelNum;

	debug_log("NetworkNodeNum: %u, LinkNum: %u, ConsumptionNodeNum: %u, HardwareLevelNum: %u\n", networkNodeNum, linkNum, consumptionNodeNum, hardwareLevelNum);

	{
		// Complexity defaultComplexity = defineDefaultComplexity(); // local variable, just print the complexity for debug, not used to optimize case according to its size
		// info_log("default complexity: %s\n", complexityString(defaultComplexity));
	}

	linkTuples.sort([](const struct LinkTuple& lhs, const struct LinkTuple& rhs) { return lhs.src < rhs.src; });

	// 2. ===============    construct graph    ===============
	Graph graph(networkNodeNum);
	graph.construct(linkTuples);
	graph.undirectify(linkTuples);
	linkTuples.clear(); // we don't need it anymore, release its memory

	// 3. ===============    display graph    ===============
	// graph.display(); // log control inside

	// 4. ===============    add minimum cost flow algorithm related components into graph    ===============
	graph.preMinimumCostFlow();

	// 5. ===============    critical part of finding a server deployment solution    ===============
	uint32_t i = 0;
	std::list<int> newServers, bestServers;
	std::vector<int> serversOutputFlux(networkNodeNum, 0);
	std::vector<int> restrictedOutputCap(networkNodeNum, 0);
	std::vector<int> gRestrictedOutputCap(networkNodeNum, 0);

	// 5.1. ==========    check whether the problen is unfeasible    ==========
	for (i = 0; i < networkNodeNum; ++i)
		newServers.push_back(i);

	// 5.1.1. =====    add the super source of all network nodes' union onto graph    =====
	graph.addSuperSource(newServers);

	// 5.1.2. =====    calculate minimum cost flow using all network nodes    =====
	long lowboundRentalFee = graph.minimumCostFlow(serversOutputFlux);

	// 5.1.3. =====    delete the super source from graph and revert the status of arc to initialization    =====
	graph.delSuperSource();

	// 5.1.4. =====    check whether no feasible scheme can be found    =====
	if (lowboundRentalFee == -1)
	{
		uncond_log("This case is unfeasible, write NA to result file.\n");

		// 5.1.5. =====    delete minimum cost flow algorithm related components from graph    =====
		graph.postMinimumCostFlow();

		// 5.1.6. =====    destruct graph    =====
		graph.destruct();

		// 5.1.7. =====    output NA to result file    =====
		outputToFile(argv[2]);

		return 0; // no need to execute rest code, thus return here
	}
	else
		newServers.clear();

	// 5.2. ==========    scan topology to obtain combinations of some carefully selected network nodes as initial scheme    ==========
	TopologyScanner topoScanner(graph);

	topoScanner.scan();
	topoScanner.recommend();

	// 5.3. ==========    deploy CDN servers on one of these combinations of network nodes    ==========
	Greedy greedy(topoScanner);

	long gMinTotalFee = INF;
	double elapsedTime = 0.0;
	std::list<int>::iterator itS;
	int LOOP = networkNodeNum < 1000 ? 3 : 1;

do {
	greedy.initialize(newServers);

	// info_log("initial servers: ");
	// printList(newServers, "");

	int loop = 1;
	long minTotalFee = INF;
	bool continueLoop = true;
	while (true)
	{
		// 5.4. ==========    add a super source onto graph    ==========
		graph.addSuperSource(newServers);

		// 5.5. ==========    execute minimum cost flow algorithm    ==========
		long rentalFee = graph.minimumCostFlow(serversOutputFlux);
		long purchaseFee = 0;
		long deploymentFee = 0;
		if (rentalFee != -1)
		{
			for (itS = newServers.begin(); itS != newServers.end(); ++itS)
				deploymentFee += deployTable[*itS];
			purchaseFee = obtainTotalPurchaseFee(serversOutputFlux, newServers.size());
		}

		// info_log("Link fee is %ld, purchase fee is %ld, deploy fee is %ld, sum up %ld.\n", rentalFee, purchaseFee, deploymentFee, rentalFee + purchaseFee + deploymentFee);

		// 5.6. ==========    a better deployment solution ?    ==========
		if (minTotalFee > rentalFee + purchaseFee + deploymentFee && rentalFee != -1)
		{
			minTotalFee = rentalFee + purchaseFee + deploymentFee;
			curBestServers = newServers;
			for (i = 0; i < networkNodeNum; ++i)
				restrictedOutputCap[i] = graph.NodeList[i].restrictedCap;
		}

		continueLoop = greedy.optimize(newServers, serversOutputFlux, rentalFee, purchaseFee, deploymentFee);

		// 5.7. ==========    time is not enough    ==========
		if (loop % 50 == 0)
			elapsedTime = procedureTimer.elapsed();
		if (loop % 1000 == 0)
			uncond_log("Time elapsed: %fs\n", elapsedTime);
		// how many loops do you want?
		if (++loop > 20000 || !continueLoop || elapsedTime > 87.0) // we have to exit loop when only 3 seconds left
		{
			uncond_log("Exit loop: %d\n", loop);
			info_log("current best servers: ");
			printList(curBestServers, "");

			// 5.7.1. =====    delete current super source from graph and revert the status of arc to initialization    =====
			graph.delSuperSource();
			if (LOOP >= 2)
			{
				if (gMinTotalFee > minTotalFee)
				{
					gMinTotalFee = minTotalFee;
					bestServers = curBestServers;
					bestRestrictedLevelNum = hardwareLevelNum;
					for (i = 0; i < networkNodeNum; ++i)
						gRestrictedOutputCap[i] = restrictedOutputCap[i];
				}

				newServers.clear();
				--hardwareLevelNum;
				uncond_log("\nhighest available server level limited to %u\n", hardwareLevelNum-1);
				topoScanner.resetRestrictedCap();
				break;
			}
			if (gMinTotalFee > minTotalFee)
			{
				gMinTotalFee = minTotalFee;
				bestServers = curBestServers;
				for (i = 0; i < networkNodeNum; ++i)
					gRestrictedOutputCap[i] = restrictedOutputCap[i];
			}
			else
				hardwareLevelNum = bestRestrictedLevelNum;

			// 5.7.2. =====    add the super source of best servers' union onto graph    =====
			for (i = 0; i < networkNodeNum; ++i)
				graph.NodeList[i].restrictedCap = gRestrictedOutputCap[i];
			graph.addSuperSource(bestServers);

			// 5.7.3. =====    calculate minimum cost flow using best servers and then record corresponding flow path    =====
			graph.minimumCostFlow(serversOutputFlux);
			graph.recordFlowPath();

			// 5.7.4. =====    in fact, it is not must to revert status and delete super source    =====
			graph.delSuperSource();
			break;
		}

		// 5.8. ==========    delete the super source from graph and revert the status of arc to initialization    ==========
		graph.delSuperSource();

		// info_log("new servers: ");
		// printList(newServers, "");
	}
}
while (--LOOP > 0);

	// 6. ===============    delete minimum cost flow algorithm related components from graph    ===============
	graph.postMinimumCostFlow();

	info_log("Minimum total fee is %ld.\n", gMinTotalFee);

	std::unordered_map<int /* deploy node */, int /* hardware level */> finalChosenLevel;
	itS = bestServers.begin();
	for (size_t l = 0; l < bestServers.size(); ++l, ++itS)
		finalChosenLevel.insert(std::pair<int, int>(*itS, fitLevel(serversOutputFlux[l])));

	/*
	{
		FILE *fd = fopen("server-level.txt", "a");
		if (fd == NULL)
		{
			error_log("Fail to open file server-level.txt.\n");
			exit(EXIT_FAILURE);
		}

		fprintf(fd, "advanced case: 0\ntotal cost: %ld\nserver level: ", minTotalFee);
		std::unordered_map<int, int>::iterator itFCL = finalChosenLevel.begin();
		for (i = 1; itFCL != finalChosenLevel.end(); ++itFCL, ++i)
		{
			fprintf(fd, "(%d,%d) ", itFCL->first, itFCL->second);
			if (i % 10 == 0)
				fprintf(fd, "\n");
		}
		fprintf(fd, "\n\n\n");

		fclose(fd);
	}
	*/

	// 7. ===============    modify flux path list to meet output requirement    ===============
	for (std::list<std::vector<int> >::iterator itfp = graph.fluxPathList.begin(); itfp != graph.fluxPathList.end(); ++itfp)
	{
		std::vector<int>& fluxPath = *itfp; // alias
		int storeFlux = fluxPath.back();
		fluxPath.pop_back();
		fluxPath.push_back(provideTable[fluxPath.back()]);
		fluxPath.push_back(storeFlux);
		fluxPath.push_back(finalChosenLevel[fluxPath.front()]);
		// printVector(fluxPath, "");
	}

	// 8. ===============    verify servers' deployment, purchase and flux paths scheme    ===============
	long totalRentalFee = 0, totalPurchaseFee = 0, totalDeploymentFee = 0;
	bool schemeOK = verifyScheme(graph, totalRentalFee, totalPurchaseFee);
	for (itS = bestServers.begin(); itS != bestServers.end(); ++itS)
		totalDeploymentFee += deployTable[*itS];
	if (schemeOK == true)
		uncond_log("Congratulations! Total rental fee is %ld, purchase fee is %ld, deployment fee is %ld, sum up %ld.\n",
			totalRentalFee, totalPurchaseFee, totalDeploymentFee, totalRentalFee + totalPurchaseFee + totalDeploymentFee);
	else
		error_log("Shit! The scheme failed to pass verification!\n");

	// 9. ===============    destruct graph    ===============
	graph.destruct();

	// 10. ===============    output flux path list to result file    ===============
	outputToFile(argv[2], graph.fluxPathList);

	/*******************************************************************
	 ********************     Time elapsing end     ********************
	 *******************************************************************/
}
