/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: graphGen.cpp                                                               *
 * @author: Xu Le                                                                    *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
// g++ -Wall -std=c++11 graphGen.cpp -o graphGen
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <vector>
#include <list>
#include <set>
#include <algorithm>
#include <functional>

#define MIN_CONNECTED_LINK_NUM      1
#define MAX_CONNECTED_LINK_NUM    100

static int MAX_LINK_BANDWIDTH = 50;
static int MAX_LINK_RENTAL_FEE = 10;
static int BASE_LEVEL_CAP = 100;
static int BASE_LEVEL_FEE = 50;
static int LOWBOUND_DEPLOY_FEE = 50;
static int UPBOUND_DEPLOY_FEE = 150;
static int LOWBOUND_DEMAND = 10;
static int UPBOUND_DEMAND = 100;
static double decayFactor = 0.6;

static unsigned int seed = 1;

static int networkNodeNum = 0;
static int linkNum = 0;
static int consumptionNodeNum = 0;
static int hardwareLevelNum = 0;

struct ServerTuple
{
	ServerTuple(int _level, int _cap, int _fee) : level(_level), cap(_cap), fee(_fee) {}

	ServerTuple(const ServerTuple& rhs) = default;
	ServerTuple& operator=(const ServerTuple& rhs) = default;

	int level;
	int cap;
	int fee;
};

struct LinkTuple
{
	LinkTuple(int _src, int _dst, int _bandwidth, int _rental) : src(_src), dst(_dst), bandwidth(_bandwidth), rental(_rental) {}

	LinkTuple(const LinkTuple& rhs) = default;
	LinkTuple& operator=(const LinkTuple& rhs) = default;

	int src;
	int dst;
	int bandwidth;
	int rental;
};

struct ComsumptionTuple
{
	ComsumptionTuple(int _cspID, int _netID, int _demand) : cspID(_cspID), netID(_netID), demand(_demand) {}

	ComsumptionTuple(const ComsumptionTuple& rhs) = default;
	ComsumptionTuple& operator=(const ComsumptionTuple& rhs) = default;
	
	int cspID;  ///< comsumptionID
	int netID;  ///< networkID
	int demand; ///< bandwidth occupy demand
};

static void printHelp()
{
	printf("Usage:\n    ./graphGen case_x.txt -n netNum -a linkNum -c cspNum -h hwLevelNum -C baseLevelCap -F baseLevelFee\n");
	printf("    [-b maxBW -r maxRentalFee -L lbDeployFee -U ubDeployFee -l lbDemand -u ubDemand -d decayFactor -s seed]\n");
	printf("An example:\n    ./graphGen case_0.txt -n 300 -a 1200 -c 60 -h 5 -C 100 -F 50 [-b 50 -r 10 -L 50 -U 150 -l 10 -u 100 -d 0.6 -s 1]\n");
}

inline int randomSrc() { return rand() % networkNodeNum; }
inline int randomDst() { return rand() % networkNodeNum; }
inline int randomAccess() { return rand() % networkNodeNum; }
inline int randomCap() { return rand() % MAX_LINK_BANDWIDTH + 1; }
inline int randomFee() { return rand() % MAX_LINK_RENTAL_FEE + 1; }
inline int randomDeploy() { return rand() % (UPBOUND_DEPLOY_FEE - LOWBOUND_DEPLOY_FEE + 1) + LOWBOUND_DEPLOY_FEE; }
inline int randomDemand() { return rand() % (UPBOUND_DEMAND - LOWBOUND_DEMAND + 1) + LOWBOUND_DEMAND; }
inline int randomCapIncrement() { return rand() % (2*BASE_LEVEL_CAP - BASE_LEVEL_CAP/2 + 1) + BASE_LEVEL_CAP/2; }
inline long uuid(int src, int dst) { return (src << 16) + dst; }

static void generateServer(std::list<struct ServerTuple>& serverTuples)
{
	int curLevel = 0, capacity = BASE_LEVEL_CAP, price = BASE_LEVEL_FEE, capacityIncrement = 0, priceIncrement = 0;
	double unitCapFee = static_cast<double>(BASE_LEVEL_FEE) / BASE_LEVEL_CAP;
	for (; curLevel < hardwareLevelNum; ++curLevel)
	{
		serverTuples.emplace_back(curLevel, capacity, price);
		unitCapFee /= decayFactor;
		do { capacityIncrement = randomCapIncrement(); } while (capacityIncrement % 10 != 0);
		priceIncrement = static_cast<double>(capacityIncrement) * unitCapFee;
		capacity += capacityIncrement;
		price += priceIncrement;
		while (price % 10 != 0)
			++price;
	}
}

static void generateTopology(std::list<struct LinkTuple>& linkTuples)
{
	int arcNum = linkNum;

	std::vector<int> outLinkNum(networkNodeNum, 0);
	std::set<long> linkUUID;

	int src = 0, dst = 0;
	int minOutLink = MIN_CONNECTED_LINK_NUM, hasAlreadyOutLink = 0;
	while (arcNum > 0)
	{
		if (hasAlreadyOutLink++ < minOutLink)
		{
			for (src = 0; src < networkNodeNum; ++src)
			{
				do { dst = randomDst(); } while (outLinkNum[dst] > MAX_CONNECTED_LINK_NUM);
				while (linkUUID.find(uuid(src, dst)) != linkUUID.end() || linkUUID.find(uuid(dst, src)) != linkUUID.end() || src == dst)
					dst = randomDst();
				linkTuples.emplace_back(src, dst, randomCap(), randomFee());
				linkUUID.insert(uuid(src, dst));
				linkUUID.insert(uuid(dst, src));
				++outLinkNum[src];
				++outLinkNum[dst];
				--arcNum;
			}
		}
		else
			break;
	}
	while (arcNum > 0)
	{
		do { src = randomSrc(); } while (outLinkNum[dst] > MAX_CONNECTED_LINK_NUM);
		do { dst = randomDst(); } while (outLinkNum[dst] > MAX_CONNECTED_LINK_NUM);
		while (linkUUID.find(uuid(src, dst)) != linkUUID.end() || linkUUID.find(uuid(dst, src)) != linkUUID.end() || src == dst)
			dst = randomDst();
		linkTuples.emplace_back(src, dst, randomCap(), randomFee());
		linkUUID.insert(uuid(src, dst));
		linkUUID.insert(uuid(dst, src));
		++outLinkNum[src];
		++outLinkNum[dst];
		--arcNum;
	}

	linkTuples.sort([](const LinkTuple& lhs, const LinkTuple& rhs) {
		return lhs.src < rhs.src;
	});
}

static void generateCspNodes(std::list<struct ComsumptionTuple>& cspTuples)
{
	std::vector<bool> hasCsp(networkNodeNum, false);

	int accessNode = 0;
	for (int i = 0; i < consumptionNodeNum; ++i)
	{
		do { accessNode = randomAccess(); } while (hasCsp[accessNode]);
		cspTuples.emplace_back(i, accessNode, randomDemand());
		hasCsp[accessNode] = true;
	}
}

static void generateCaseFile(const char *filename, std::list<struct ServerTuple>& serverTuples, std::list<struct LinkTuple>& linkTuples, std::list<struct ComsumptionTuple>& cspTuples)
{
	FILE *fd = fopen(filename, "w");
	if (fd == NULL)
	{
		fprintf(stderr, "cannot open %s!\n", filename);
		exit(EXIT_FAILURE);
	}

	fprintf(fd, "%d %d %d\n\n", networkNodeNum, linkNum, consumptionNodeNum);

	std::list<struct ServerTuple>::iterator itS = serverTuples.begin();
	while (hardwareLevelNum-- > 0)
	{
		fprintf(fd, "%d %d %d\n", itS->level, itS->cap, itS->fee);
		++itS;
	}

	fprintf(fd, "\n");

	int deployFee;
	for (int id = 0; id < networkNodeNum; ++id)
	{
		do { deployFee = randomDeploy(); } while (deployFee % 10 != 0);
		while (deployFee % 100 != 0)
			++deployFee;
		fprintf(fd, "%d %d\n", id, deployFee);
	}

	fprintf(fd, "\n");

	std::list<struct LinkTuple>::iterator itLink = linkTuples.begin();
	while (linkNum-- > 0)
	{
		fprintf(fd, "%d %d %d %d\n", itLink->src, itLink->dst, itLink->bandwidth, itLink->rental);
		++itLink;
	}

	fprintf(fd, "\n");

	std::list<struct ComsumptionTuple>::iterator itCsp = cspTuples.begin();
	while (consumptionNodeNum-- > 0)
	{
		fprintf(fd, "%d %d %d\n", itCsp->cspID, itCsp->netID, itCsp->demand);
		++itCsp;
	}

	fclose(fd);
}

int main(int argc, char *argv[])
{
	if (argc < 10 || strcmp(argv[1], "-?") == 0 || strcmp(argv[1], "--help") == 0)
	{
		printHelp();
		exit(EXIT_FAILURE);
	}

	{
		std::string touchCmd("touch ");
		touchCmd.append(argv[1]);
		system(touchCmd.c_str());
	}

	{
		networkNodeNum = atoi(argv[3]);
		linkNum = atoi(argv[5]);
		consumptionNodeNum = atoi(argv[7]);
		hardwareLevelNum = atoi(argv[9]);
		BASE_LEVEL_CAP = atoi(argv[11]);
		BASE_LEVEL_FEE = atoi(argv[13]);
		if (argc > 14)
			MAX_LINK_BANDWIDTH = atoi(argv[15]);
		if (argc > 16)
			MAX_LINK_RENTAL_FEE = atoi(argv[17]);
		if (argc > 18)
			LOWBOUND_DEPLOY_FEE = atoi(argv[19]);
		if (argc > 20)
			UPBOUND_DEPLOY_FEE = atoi(argv[21]);
		if (argc > 22)
			LOWBOUND_DEMAND = atoi(argv[23]);
		if (argc > 24)
			UPBOUND_DEMAND = atoi(argv[25]);
		if (argc > 26)
			decayFactor = atof(argv[27]);
		if (argc > 28)
			seed = static_cast<unsigned int>(atoi(argv[29]));
	}

	srand(seed);

	std::list<struct ServerTuple> serverTuples;
	std::list<struct LinkTuple> linkTuples;
	std::list<struct ComsumptionTuple> cspTuples;

	generateServer(serverTuples);
	generateTopology(linkTuples);
	generateCspNodes(cspTuples);
	generateCaseFile(argv[1], serverTuples, linkTuples, cspTuples);

	return 0;
}

